"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gzip
import os
import pickle
from typing import List, Optional, Tuple

import cv2
import numpy
from PyQt6 import QtWidgets, QtGui, QtCore
import scipy.io
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC

from pyjamas.external import pascal_voc_io
from pyjamas.pjscore import PyJAMAS
from pyjamas.pjscore import undo_modes
from pyjamas.rcallbacks.rcallback import RCallback
from pyjamas.rimage.rimml.batchclassifier import BatchClassifier
from pyjamas.rimage.rimml.batchneuralnet import BatchNeuralNet
from pyjamas.rimage.rimml.rimclassifier import rimclassifier
from pyjamas.rimage.rimutils import rimutils
from pyjamas.rimage.rimml.rimlr import lr
from pyjamas.rimage.rimml.rimsvm import svm
from pyjamas.rimage.rimml.rimunet import UNet
from pyjamas.rutils import RUtils
from pyjamas.rimage.rimml.classifier_types import classifier_types
from pyjamas.dialogs.classifiertype import ClassifierTypeDialog
from pyjamas.rimage.rimml.batchrecurrentneuralnet import BatchRecurrentNeuralNet
from pyjamas.rimage.rimml.rimrescunet import ReSCUNet


class RCBIO(RCallback):
    FILENAME_BASE: str = "cell_"
    FILENAME_FIDUCIAL_LENGTH: int = 5
    DEFAULT_PICKLE_PROTOCOL: int = RUtils.DEFAULT_PICKLE_PROTOCOL

    def cbLoadTimeSeries(self, filename: Optional[str] = None, channel_method: Optional[rimutils.ChannelMethod] = None,
                         index: int = -1, message: str = "Load image ...") -> bool:
        """
        Load an image from file using rimutils.read_stack.

        :param filename: file to open.
        :param channel_method: for a multi-channel sequence, determines whether to open a channel or a slice.
        :param index: index of the channel or slice to open (with origin at 0).
        :param message: string to display as the dialog box title.
        :return: True if the image was loaded with no problems, False otherwise.
        """

        # Get file name.
        if filename is None or filename is False:
            fname = QtWidgets.QFileDialog.getOpenFileName(None, message, self.pjs.cwd,
                                                          filter='All files (*)')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return False

        _, ext = os.path.splitext(filename)

        # Read image.
        theimage = rimutils.read_stack(filename, channel_method, index)
        if theimage is not None:
            self.pjs.slices = theimage
        else:
            self.pjs.statusbar.showMessage("That does not look like an image.")
            return False

        # Clear annotation undo stack.
        self.pjs.undo_stack.clear()

        # Store current working directory.
        self.pjs.cwd = os.path.dirname(filename)  # Path of loaded image.
        self.pjs.filename = filename

        # Initialize image and display.
        self.pjs.initImage()

        # Clear and reset classifier object arrays so that the classifier can be applied without reloading from file
        self.pjs.batch_classifier.reset_objects(self.pjs.n_frames)

        return True

    def cbLoadArray(self, image: numpy.ndarray) -> bool:
        """
        Load a numpy matrix as an image.

        :param image: the array to open.
        :return: True if the image was loaded with no problems, False otherwise.
        """
        if image is False or image is None:
            return False

        self.pjs.slices = image

        self.pjs.filename = 'animage'

        self.pjs.initImage()

        # Clear and reset classifier object arrays so that the classifier can be applied without reloading from file
        self.pjs.batch_classifier.reset_objects(self.pjs.n_frames)

    def cbLoadAnnotations(self, filenames: Optional[List[str]] = None, image_file: Optional[str] = None,
                          replace: bool = True, message: str = "Load annotations ...") -> bool:  # Handle IO errors
        """
        Loads PyJAMAS annotation files.

        :param filenames: paths to the files containing the annotations.
        :param image_file: path to an image to be loaded with the annotation file. None if no image is to be loaded. '' to create an empty image.
        :param replace: True if loaded annotations should replace existing ones, False if this is an additive load.
        :param message: string to display as the dialog box title.
        :return: True if the annotation file was loaded, False otherwise.
        """

        # Get file name.
        if filenames is None or filenames is False or filenames == []: # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            dialog = QtWidgets.QFileDialog(None, message)
            dialog.setFileMode(QtWidgets.QFileDialog.FileMode.ExistingFiles)
            dialog.setDirectory(self.pjs.cwd)
            dialog.setNameFilter('PyJAMAS data (*' + PyJAMAS.data_extension + ')')
            dialog.exec()

            filenames = dialog.selectedFiles()
            if filenames is None or filenames == [] or filenames[0] == '':
                return False

        filenames.sort(key=RUtils.natural_sort)

        if image_file is not False and image_file is not None and image_file != '':
            self.pjs.io.cbLoadTimeSeries(image_file)

        # # Open file names and read annotations.
        fh = None
        maxid: int = 0

        if not replace:
            [allfiducials, allqpolygonfs, allpolyline_ids] = self.pjs.copy_annotations()
            allpolylines = RUtils.qpolygonfs2coordinatelists(allqpolygonfs)

            maxid = max([0] + [max(slice_ids) for slice_ids in allpolyline_ids if
                               slice_ids])  # the [0] ensures no error if there are no annotations.

        else:
            allfiducials = [[] for _ in range(self.pjs.n_frames)]
            allpolylines = [[] for _ in range(self.pjs.n_frames)]
            allpolyline_ids = [[] for _ in range(self.pjs.n_frames)]

        curfiducials: List[List[int]] = None
        curpolylines: List[List[int]] = None
        curpolyline_ids: List[List[int]] = None

        for fname in filenames:
            try:
                fh = gzip.open(fname, "rb")
                curfiducials = pickle.load(fh)
                curpolylines = pickle.load(fh)
                curpolyline_ids = pickle.load(fh)

            except (IOError, OSError) as ex:
                if fh is not None:
                    fh.close()
                print(ex)
                return False

            except EOFError as ex:
                if not curfiducials:
                    curfiducials = [[] for _ in range(self.pjs.n_frames)]
                    curpolylines = [[] for _ in range(self.pjs.n_frames)]
                    curpolyline_ids = [[] for _ in range(self.pjs.n_frames)]

                elif not curpolylines:
                    curpolylines = [[] for _ in range(self.pjs.n_frames)]
                    curpolyline_ids = [[] for _ in range(self.pjs.n_frames)]

                # this takes care of old annotation files with no polyline_ids.
                elif not curpolyline_ids:
                    curpolyline_ids = [[] for _ in range(len(curpolylines))]

                    for slice_index, slice_polylines in enumerate(curpolylines):
                        for thepolyline_index, thepolyline in enumerate(slice_polylines):
                            if thepolyline:
                                curpolyline_ids[slice_index].append(thepolyline_index + 1)

            for i in range(len(allfiducials)):
                allfiducials[i].extend(curfiducials[i])

            for i in range(len(allpolylines)):
                allpolylines[i].extend([apolylinelist for apolylinelist in curpolylines[i]])
                allpolyline_ids[i].extend((numpy.asarray(curpolyline_ids[i]) + maxid).tolist())

            max_id_list = [max(slice_ids) for slice_ids in curpolyline_ids if slice_ids]

            # this takes care of the scenario with no polylines defined on any slice
            if max_id_list:
                themaxid = max(max_id_list)
                maxid = max(maxid, themaxid)

        if image_file == '':  # Create a blank image.
            # Find maxx, maxy, and maxz, and create a numpy.ndarray with those dimensions + 1.
            maxz = len(allfiducials)
            maxx = 0
            maxy = 0

            # Fiducials and polylines are stored as (x, y) coordinates.
            # numpy.ndarrays are created with (slices, rows, cols).
            for slice_fiducials in allfiducials:
                for afiducial in slice_fiducials:
                    maxx = max(maxx, afiducial[0])
                    maxy = max(maxy, afiducial[1])

            for slice_polylines in allpolylines:
                for apolyline in slice_polylines:
                    for afiducial in apolyline:
                        maxx = max(maxx, afiducial[0])
                        maxy = max(maxy, afiducial[1])

            virtual_image = numpy.zeros((maxz, int(maxy + 1), int(maxx + 1)), dtype=int)
            self.pjs.io.cbLoadArray(virtual_image)

        # elif image_file is not False and image_file is not None:
        #    self.pjs.io.cbLoadTimeSeries(image_file)

        self.pjs.annotations.cbDeleteAllAnn()

        self.pjs.undo_stack.push(
            {'changetype': undo_modes.MULTI, 'frame': (0, self.pjs.n_frames), 'index': (True, True),
             'details': self.pjs.copy_annotations()})

        for slice_index, slice_fiducials in enumerate(allfiducials):
            for afiducial in slice_fiducials:
                self.pjs.addFiducial(int(afiducial[0]), int(afiducial[1]), slice_index, paint=False)

        for i, theframepolylines in enumerate(allpolylines):
            if i < self.pjs.n_frames:  # this allows opening annotation files with extra time poinnts.
                for j, thepolyline in enumerate(theframepolylines):
                    if thepolyline != []:
                        if self.pjs.close_all_polylines and thepolyline[0] != thepolyline[-1]:
                            thepolyline.append(thepolyline[0])
                        self.pjs.addPolyline(thepolyline, i, paint=False)
                        # self.pjs.polylines[i].append(thepolyline)
                        self.pjs.polyline_ids[i][-1] = allpolyline_ids[i][j]

        # self.pjs.image.cbZoom(0)
        self.pjs.repaint()

        # Modify current path.
        self.pjs.cwd = os.path.dirname(filenames[0])

        return True

    def cbSaveTimeSeries(self, filename: Optional[str] = None, message: str = "Save time series ...") -> bool:
        '''
        Save the current image (pjs.slices) as a grayscale, multi-page TIFF.

        :param filename: path to the destination file.
        :param message: string to display as the dialog box title.
        :return: True if the image was saved with no problems, False otherwise.
        '''

        # Get file name.
        if filename == '' or filename is False or filename is None:
            fname: tuple = QtWidgets.QFileDialog.getSaveFileName(None, message, self.pjs.cwd,
                                                          filter='TIFF files (*.tif *.tiff)')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
        if filename == '':
            return False

        # Save image.
        rimutils.write_stack(filename, self.pjs.slices)
        self.pjs.statusbar.showMessage('Image saved: ' + filename)

        self.pjs.cwd = os.path.dirname(filename)  # Path of loaded image.
        self.pjs.filename = filename

        return True

    def cbExportROIAndMasks(self, polyline: Optional[numpy.ndarray] = None, margin_size: int = 0) -> bool:
        """
        Save the image within an ROI, plus a second image containing a binary mask of any polylines within the ROI.

        :param polyline: ndarray with two columns containing the x, y coordinates of the ROI; if using a tracked polyline, ndarray with one element corresponding to the polyline index; if not provided, the function will prompt the user to select a polyline.
        :param margin_size: margin size in pixels for cropping around the polyline
        :return: True if the image is cropped, False otherwise.

        """
        if (polyline is None) | (polyline == []):
            thepolylines = self.pjs.polylines[self.pjs.curslice]

            if thepolylines == [] or thepolylines[0] == []:
                return False

            # prompt user to select polyline
            self.pjs.statusbar.showMessage('Select polygons.')
            self.pjs.annotation_mode = PyJAMAS.select_polyline_exportroi

            return True

        minx, miny, maxx, maxy = self.get_coordinate_bounds(coordinates=polyline, margin_size=margin_size)

        # Where to save?
        filenameext = self.generate_ROI_filename((minx, maxx), (miny, maxy), (self.pjs.curslice, self.pjs.curslice),
                                                 PyJAMAS.image_extensions[0], relative=True)
        filename, _ = os.path.splitext(filenameext)
        path_to_image: str = os.path.join(self.pjs.cwd, filename, "image", filenameext)
        path_to_mask: str = os.path.join(self.pjs.cwd, filename, "mask", filenameext)
        cwd = self.pjs.cwd

        # Save image.
        self.pjs.io.cbSaveROI(path_to_image, (minx, maxx), (miny, maxy), (self.pjs.curslice, self.pjs.curslice))

        # Export mask.
        self.pjs.io.cbExportCurrentAnnotationsBinaryImage(path_to_mask, numpy.array([[minx, miny], [maxx, maxy]]))

        # Important, as cbSaveROI and cbExportCurrentAnnotationsBinaryImage may affect the cwd.
        self.pjs.cwd = cwd

        return True

    def cbSaveROI(self, filename: Optional[str] = None, x_range: Tuple[int, int] = False,
                  y_range: Tuple[int, int] = False, z_range: Tuple[int, int] = False,
                  message: str = "Save time series ...") -> bool:
        """
        Save a subregion of the image currently open (pjs.slices).

        :param filename: '' for automated naming based on coordinates. If None, a dialog will open. Make sure to provide an extension (defaults to PyJAMAS.image_extensions[0]).
        :param x_range: tuple containing the min and max X values to save. If False, take the coordinates of the first polyline defined on the image. If no polygons, use the entire image width.
        :param y_range: tuple containing the min and max Y values to save. If False, take the coordinates of the first polyline defined on the image. If no polygons, use the entire image height.
        :param z_range: tuple containing the min and max Z values to save. If False, use just the current Z.
        :param message: string to display as the dialog box title.
        :return: True if the image was properly saved, False otherwise.
        """

        # Check X, Y, and Z parameters or assign default values.
        if z_range is False:
            z_range = tuple([self.pjs.curslice, self.pjs.curslice])

        if x_range is False:
            thepolylines = self.pjs.polylines[self.pjs.curslice]
            if thepolylines == [] or thepolylines[0] == []:
                return False

            thepolyline = self.pjs.polylines[self.pjs.curslice][0].boundingRect()
            x_range = tuple([int(thepolyline.x()), int(thepolyline.x()+thepolyline.width())])
        if y_range is False and thepolylines != [] and thepolylines[0] != []:
            thepolylines = self.pjs.polylines[self.pjs.curslice]
            if thepolylines == [] or thepolylines[0] == []:
                return False

            thepolyline = self.pjs.polylines[self.pjs.curslice][0].boundingRect()
            y_range = tuple([int(thepolyline.y()), int(thepolyline.y() + thepolyline.height())])
        # Get file name.
        if filename is None or filename is False:
            fname: tuple = QtWidgets.QFileDialog.getSaveFileName(None, message, self.pjs.cwd,
                                                          filter='TIFF files (*.tif *.tiff)')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return False

        elif filename == '':
            filename = self.generate_ROI_filename(x_range, y_range, z_range, PyJAMAS.image_extensions[0])
        # Make sure the file name has an extension (this is important to determine the image format in rimutils.write_stack.
        _, fext = os.path.splitext(filename)
        if fext == '':
            filename = RUtils.set_extension(filename, PyJAMAS.image_extensions[0])

        # Save image.
        rimutils.write_stack(filename,
                             self.pjs.slices[z_range[0]:z_range[1]+1, y_range[0]:y_range[1]+1, x_range[0]:x_range[1]+1])
        self.pjs.statusbar.showMessage('Image saved: ' + filename)

        self.pjs.cwd = os.path.dirname(filename)  # Path of loaded image.

        return True

    def cbSaveDisplay(self, filename: Optional[str] = None, message: str = "Save display ...") -> bool:
        """
        Save the PyJAMAS display, including annotations, as a colour image.

        :param filename: '' for automated naming based on the current image name (pjs.filename). If not provided, a dialog will open.
        :param message: string to display as the dialog box title.
        :return: True if the display was properly saved, False otherwise.
        """

        # Get file name.
        if filename is False or filename is None:
            fname: tuple = QtWidgets.QFileDialog.getSaveFileName(None, message, self.pjs.cwd,
                                                          filter='TIFF files (*.tif *.tiff)')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return False

        elif filename == '':
            _, fname = os.path.split(self.pjs.filename)
            filename = os.path.join(self.pjs.cwd, fname)

        # Save image.
        self.capture_display(filename)

        self.pjs.statusbar.showMessage('Display saved: ' + filename)

        self.pjs.cwd = os.path.dirname(filename)  # Path of loaded image.

        return True

    def cbExportMovie(self, filename: Optional[str] = None, message: str = "Export movie with annotations ...") -> bool:
        """
        Save an avi file displaying the movie currently open and its annotations.
        The number of frames per second is determined by pjs.fps.

        :param filename: '' for automated naming based on the current image name (pjs.filename). If not provided, a dialog will open.
        :param message: string to display as the dialog box title.
        :return: True if the display was properly saved, False otherwise.
        """

        # Get file name.
        if filename is False or filename is None:
            fname: tuple = QtWidgets.QFileDialog.getSaveFileName(None, message, self.pjs.cwd,
                                                          filter='AVI (*.avi)')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return False

        elif filename == '':
            _, fname = os.path.split(self.pjs.filename)
            filename = os.path.join(self.pjs.cwd, fname)

        # Save image.
        self.export_movie(filename)

        self.pjs.statusbar.showMessage('Movie exported: ' + filename)

        self.pjs.cwd = os.path.dirname(filename)  # Path of loaded image.

        return True

    def capture_display(self, filename: Optional[str] = None) -> QtGui.QImage:

        #pix_map: QtGui.QPixmap = self.pjs.gView.grab(self.pjs.gView.sceneRect().toRect())
        #pix_map: QtGui.QPixmap = self.pjs.gView.grab(QRect(0, 0, int(self.pjs.width * self.pjs.zoom_factors[self.pjs.zoom_index]), int(self.pjs.height * self.pjs.zoom_factors[self.pjs.zoom_index])))
        #pix_map: QtGui.QPixmap = self.pjs.gView.grab()
        #pix_map = pix_map.scaled(pix_map.width(), pix_map.height())

        self.pjs.gScene.clearSelection()
        self.pjs.gScene.setSceneRect(self.pjs.gScene.itemsBoundingRect())
        image: QtGui.QImage = QtGui.QImage(self.pjs.gScene.sceneRect().size().toSize(), QtGui.QImage.Format.Format_ARGB32)
        image.fill(QtCore.Qt.GlobalColor.transparent)
        painter: QtGui.QPainter = QtGui.QPainter(image)
        self.pjs.gScene.render(painter)

        if filename and filename is not None and filename != '':
            image.save(filename)

        return image

    def export_movie(self, filename: Optional[str] = None) -> list:
        curslice = self.pjs.curslice

        qimage_list = []

        for i in range(self.pjs.n_frames):
            self.pjs.image.cbGoTo(i)
            qimage_list.append(self.capture_display())


        self.pjs.image.cbGoTo(curslice)

        if filename and filename is not None and filename != '':
            fourcc = cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')

            out = cv2.VideoWriter(filename, fourcc, self.pjs.fps, (self.pjs.width, self.pjs.height), True)

            for i in range(self.pjs.n_frames):
                a = rimutils.qimage2ndarray(qimage_list[i])

                # In some cases, particularly colour images with annotations, qimage2ndarray adds 2-3 rows of black
                # pixels at the top and left sides of the image. Here we remove them.
                h0: int = a.shape[0]-self.pjs.height
                w0: int = a.shape[1]-self.pjs.width
                gray_3c = cv2.merge([a[h0:, w0:, 0], a[h0:, w0:,  1], a[h0:, w0:, 2]])
                out.write(gray_3c.astype('uint8'))

            out.release()

        return qimage_list

    def cbSaveAnnotations(self, filename: Optional[str] = None, polylines: Optional[List] = None,
                          polyline_ids: Optional[List] = None,
                          fiducials: Optional[List] = None, pickle_protocol: int = DEFAULT_PICKLE_PROTOCOL,
                          message: str = "Save annotations ...") -> bool:  # Handle IO errors.
        """
        Save annotations on the current image in pjs format (pickled). These annotations CAN BE opened by PyJAMAS.

        :param filename: '' for automated naming based on the current image name (pjs.filename). If not provided, a dialog will open.
        :param polylines: list of polylines to save (one slice per time point, one QtGui.QPolygonF per polyline).
        :param polyline_ids: list of polylines ids to save (one slice per time point, one integer per polyline).
        :param fiducials: list of fiducials to save (one slice per time point, one row with [x, y] coordinates per fiducial (int)).
        :param pickle_protocol: integer indicating the pickle protocol to use for saving (defaults to RCBIO.DEFAULT_PICKLE_PROTOCOL).
        :param message: string to display as the dialog box title.
        :return: True if the annotations were saved, False otherwise.
        """

        # Get file name.
        if filename is None or filename is False or filename == '': # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            fname = QtWidgets.QFileDialog.getSaveFileName(None, message, self.pjs.cwd,
                                                          filter='PyJAMAS data (*' + PyJAMAS.data_extension + ')')  # fname[0] is the full filename, fname[1] is the filter used.

            # If cancel ...
            filename = fname[0]
            if filename == '':
                return False

        elif filename == '':
            _, fname = os.path.split(self.pjs.filename)
            filename = os.path.join(self.pjs.cwd, fname)

        self.pjs.cwd = os.path.dirname(filename)

        if filename[-4:] != PyJAMAS.data_extension:
            filename = RUtils.set_extension(filename, PyJAMAS.data_extension)

        if not polylines:
            polylines = self.pjs.polylines

        if not polyline_ids:
            polyline_ids = self.pjs.polyline_ids

        if not fiducials:
            fiducials = self.pjs.fiducials

        # Prepare polygons to be stored: pickle does not support QPolygonF, so we convert the polygons to a list.
        # We store the polygons as QPolygonF because QGraphicsScene does have an addPolygon method that takes a
        # QPolygonF as the parameter.
        polyline_list = RUtils.qpolygonfs2coordinatelists(polylines)

        # Open file for writing.
        fh = None

        try:
            fh = gzip.open(filename, "wb")
            pickle.dump(fiducials, fh, protocol=pickle_protocol)
            pickle.dump(polyline_list, fh, protocol=pickle_protocol)
            pickle.dump(polyline_ids, fh, protocol=pickle_protocol)

        except (IOError, OSError) as ex:
            if fh is not None:
                fh.close()

            print(ex)
            return False

        self.pjs.cwd = os.path.dirname(filename)

        self.pjs.statusbar.showMessage(f"Saved {filename}.")

        return True

    def cbExportCurrentAnnotationsBinaryImage(self, filename: Optional[str] = None,
                                              polyline: Optional[numpy.ndarray] = None,
                                              message: str = "Save annotations ...") -> bool:
        """
        Save current slice annotations as a binary image displaying one mask per object.

        :param filename: '' for automated naming based on the current image name (pjs.filename). If not provided, a dialog will open.
        :param polyline: ndarray with two columns containing the x, y coordinates of the ROI; if None or empty array, use the entire image as ROI.
        :param message: string to display as the dialog box title.
        :return:
        """
        # Get file name.
        if filename is None or filename is False:  # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            fname = QtWidgets.QFileDialog.getSaveFileName(None, message,
                                                          self.pjs.cwd,
                                                          filter='image file (*' + PyJAMAS.image_extensions[0] + ')')  # fname[0] is the full filename, fname[1] is the filter used.

            # If cancel ...
            filename = fname[0]
            if filename == '':
                return False

        elif filename == '':
            _, fname = os.path.split(self.pjs.filename)
            filename = os.path.join(self.pjs.cwd, fname)

        self.pjs.cwd = os.path.dirname(filename)

        if filename[-4:] not in PyJAMAS.image_extensions:
            filename = RUtils.set_extension(filename, PyJAMAS.image_extensions[0])

        if (polyline is None) | (polyline == []):
            minx = 0
            miny = 0
            maxx = self.pjs.width - 1
            maxy = self.pjs.height - 1
        else:
            minx, miny, maxx, maxy = self.get_coordinate_bounds(coordinates=polyline)

        # Transform annotations to export only the ones within the ROI (but not the ROI itself!).
        transform = QtGui.QTransform()
        transform.translate(-minx, -miny)

        # Transform using a one pixel less wide and less tall size such that the polyline that determines the ROI
        # (if there was one) is not included in the regions within the image.
        _, new_polylines, _ = self.transform_annotations(transform, (maxx - minx - 1, maxy - miny - 1))

        thepolylines = new_polylines[self.pjs.curslice]

        mask_image = rimutils.mask_from_polylines((maxy - miny + 1, maxx - minx + 1), thepolylines, self.pjs.brush_size)
        rimutils.write_stack(filename, mask_image)

        self.pjs.statusbar.showMessage(f"Saved {filename}.")

        return True

    def cbSaveClassifier(self, filename: Optional[str] = None,
                         theclassifier: Optional[rimclassifier] = None,
                         message: str = "Save classifier ...") -> bool:  # Handle IO errors.
        """
        Save current classifier (pickled).

        :param filename: '' for automated naming based on the current image name (pjs.filename). If not provided, a dialog will open.
        :param theclassifier: True if the classifier saved, False otherwise.
        :param message: string to display as the dialog box title.
        :return: True if the classifier was saved, False otherwise.
        """

        if theclassifier is None or theclassifier is False:
            if self.pjs.batch_classifier.image_classifier is None:
                return False
            else:
                theclassifier = self.pjs.batch_classifier.image_classifier

        # Get file name.
        if filename is None or filename is False:  # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            fname = QtWidgets.QFileDialog.getSaveFileName(None, message, self.pjs.cwd,
                                                          filter='PyJAMAS classifier (*' + PyJAMAS.classifier_extension + ')')  # fname[0] is the full filename, fname[1] is the filter used.

            # If cancel ...
            filename = fname[0]
            if filename == '':
                return False

        elif filename == '':
            _, fname = os.path.split(self.pjs.filename)
            filename = os.path.join(self.pjs.cwd, fname)

        if filename[-4:] != PyJAMAS.classifier_extension:
            filename = RUtils.set_extension(filename, PyJAMAS.classifier_extension)

        self.pjs.cwd = os.path.dirname(filename)

        theclassifier.save(filename)

        self.pjs.statusbar.showMessage(f'Saved classifier {filename}.')

        return True

    def cbLoadClassifier(self, filename: Optional[str] = None,
                         message: str = "Load classifier ...") -> bool:  # Handle IO errors
        """
        Load a classifier from disk.

        :param filename: name of the file containing a pickled classifier to be loaded.
        :param message: string to display as the dialog box title.
        :return: True if the classifier was loaded, False otherwise.
        """

        filename = self.load_classifier(filename, message)

        if filename is None:
            return False

        # Modify current path.
        self.pjs.cwd = os.path.dirname(filename)

        self.pjs.statusbar.showMessage(f"Classifier {filename} loaded.")

        return True

    def load_classifier(self, filename: Optional[str] = None, message: str = "Load classifier ...") -> str:
        # Get file name.
        if filename is None or filename is False:
            fname = QtWidgets.QFileDialog.getOpenFileName(None, message, self.pjs.cwd,
                                                          filter='PyJAMAS classifier (*' + PyJAMAS.classifier_extension + ')')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return None

        # Open file name and read classifier.
        fh = None
        theparameters: dict = None

        try:
            fh = gzip.open(filename, "rb")
            theparameters = pickle.load(fh)
            fh.close()

        except (IOError, OSError) as ex:
            if fh is not None:
                fh.close()

            print(ex)
            return None

        if theparameters is None:
            return None

        theclassifier = theparameters.get('classifier', None)
        classifier_type = theparameters.get('classifier_type', None)

        classifier_identities = set(thetype.value for thetype in classifier_types)
        loading = True
        while loading:  # Ask for classifier type until successfully built or cancelled
            try:
                if classifier_type == classifier_types.SUPPORT_VECTOR_MACHINE.value and type(theclassifier) == SVC:
                    self.pjs.batch_classifier = BatchClassifier(self.pjs.n_frames, svm(theparameters))
                elif classifier_type == classifier_types.LOGISTIC_REGRESSION.value and type(theclassifier) == LogisticRegression:
                    self.pjs.batch_classifier = BatchClassifier(self.pjs.n_frames, lr(theparameters))
                elif classifier_type == classifier_types.UNET.value and type(theclassifier) == list:
                    self.pjs.batch_classifier = BatchNeuralNet(self.pjs.n_frames, UNet(theparameters))
                elif classifier_type == classifier_types.RESCUNET.value and type(theclassifier) == list:
                    concatenation_level = theparameters.get('concatenation_level', None)
                    if concatenation_level is None:  # compatibility with older classifiers without concatenation_level
                        build_success = False  # Iterate over possible values to find proper construction
                        concatenation_level = 0
                        while not build_success and concatenation_level <= 4:
                            try:
                                theparameters['concatenation_level'] = concatenation_level
                                self.pjs.batch_classifier = BatchRecurrentNeuralNet(self.pjs.n_frames, ReSCUNet(theparameters))
                                build_success = True
                            except:
                                concatenation_level += 1
                        if not build_success:
                            raise TypeError("Wrong classifier type.")
                    else:
                        self.pjs.batch_classifier = BatchRecurrentNeuralNet(self.pjs.n_frames, ReSCUNet(theparameters))
                else:
                    raise TypeError("Wrong classifier type.")
                loading = False  # Passed through if/else's without error so the classifier was loaded, exit the loop

            except:  # Classifier type non-existent, unknown, or incorrect
                dialog = QtWidgets.QDialog()
                ui = ClassifierTypeDialog()
                ui.setupUi(dialog, classifier_type=classifier_types.UNKNOWN.value)
                dialog.exec()
                dialog.show()
                continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted
                classifier_type = ui.parameters()
                dialog.close()
                if not continue_flag:
                    self.pjs.statusbar.showMessage("Unknown classifier type.")
                    loading = False  # User cancelled loading, exit loop and return None as filename
                    filename = None

        return filename

    def cbExportPolylineAnnotations(self, folder_name: Optional[str] = None) -> bool:
        """
        Set PyJAMAS' annotation mode such that when a fiducial is clicked on, its containing polyline -over time- is stored in a new PyJAMAS annotation file (pickled).

        :param folder_name: folder where annotation files will be stored.
        :return: True if the annotation mode was changed, False otherwise.
        """
        if folder_name is not None and folder_name is not False:  # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            self.pjs.cwd = folder_name

        self.pjs.annotation_mode = PyJAMAS.export_fiducial_polyline
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.CursorShape.CrossCursor)

        return True

    def export_polyline_annotations(self, x: int, y: int) -> bool:
        # Make sure the fiducial is in the list.
        assert [x, y] in self.pjs.fiducials[self.pjs.curslice], "Fiducial not found!"

        # Identify the fiducial.
        fiducial_index: int = self.pjs.fiducials[self.pjs.curslice].index([x, y])

        # List of polygons and fiducials for new annotation file.
        fiducial_list: list = [[] for _ in range(self.pjs.n_frames)]
        polyline_list: list = [[] for _ in range(self.pjs.n_frames)]
        polyline_id_list: list = [[] for _ in range(self.pjs.n_frames)]

        # If no polyline encloses the fiducial, return False.
        found_one_polyline: bool = False

        # In every image:
        for slice in range(self.pjs.n_frames):
            # If there are enough fiducials and any polylines.
            if self.pjs.fiducials[slice] and fiducial_index < len(self.pjs.fiducials[slice]) and self.pjs.polylines[slice]:
                # Store the fiducial.
                fiducial_list[slice].append(self.pjs.fiducials[slice][fiducial_index])

                # Find the first polygon that contains the fiducial and store it as well.
                thefiducial: list = self.pjs.fiducials[slice][fiducial_index]
                thepolylines: list = [one_polyline for one_polyline in self.pjs.polylines[slice]]

                for index_polyline, one_polyline in enumerate(thepolylines):
                    if one_polyline.containsPoint(QtCore.QPointF(thefiducial[0], thefiducial[1]), QtCore.Qt.FillRule.OddEvenFill):
                        polyline_list[slice].append(self.pjs.polylines[slice][index_polyline])
                        polyline_id_list[slice].append(self.pjs.polyline_ids[slice][index_polyline])
                        found_one_polyline = True
                        break

        # Save new annotation file.
        if found_one_polyline:
            filename: str = os.path.join(self.pjs.cwd, self.FILENAME_BASE + str(fiducial_index+1).zfill(self.FILENAME_FIDUCIAL_LENGTH))
            self.pjs.io.cbSaveAnnotations(filename, polyline_list, polyline_id_list, fiducial_list)

        return True

    def export_all_polyline_annotations(self, slice_index: int = 0) -> bool:
        if slice_index < 0 or slice_index >= self.pjs.slices.shape[0]:
            return False

        for afidu in self.pjs.fiducials[slice_index]:
            self.export_polyline_annotations(afidu[0], afidu[1])

        return True

    def cbExportAllPolylineAnnotations(self, folder_name: Optional[str] = None,
                                       message: str = "Export files to folder ...") -> bool:
        """
        Exports each polyline in the image containing a fiducial into a single PyJAMAS annotation file - including polylines containing the same fiducial in other slices.

        :param folder_name: folder where annotation files will be stored.
        :param message: string to display as the dialog box title.
        :return: True if the export process completed normally, False otherwise.
        """
        if folder_name == '' or folder_name is False or folder_name is None: # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            folder_name = QtWidgets.QFileDialog.getExistingDirectory(None, message, self.pjs.cwd)

        # If cancel ...
        if folder_name == '':
            return False

        if os.path.exists(folder_name):
            self.pjs.cwd = os.path.abspath(folder_name)
            self.export_all_polyline_annotations(self.pjs.curslice)

            self.pjs.statusbar.showMessage(f"Done!")

            return True

        else:
            return False

    def cbImportSIESTAAnnotations(self, filenames: Optional[List[str]] = None,
                                  image_file: Optional[str] = None, replace: bool = True,
                                  message: str = "Load annotations ...") -> bool:
        """
        Read Matlab-based annotations into PyJAMAS..

        :param filenames: path to the files that will contain the Matlab-based annotations.
        :param image_file: path to an image that can be loaded with the Matlab-based annotations. '' to use an empty image.
        :param replace: True if loaded annotations should replace existing ones, False if this is an additive load.
        :param message: string to display as the dialog box title.
        :return: True if the annotation file was properly loaded, False otherwise.
        """
        # Get file name.
        if filenames is None or filenames is False or filenames == []: # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            dialog = QtWidgets.QFileDialog(None, message)
            dialog.setFileMode(QtWidgets.QFileDialog.FileMode.ExistingFiles)
            dialog.setDirectory(self.pjs.cwd)
            dialog.setNameFilter('SIESTA annotations (*.mat)')
            dialog.exec()

            filenames = dialog.selectedFiles()
            if filenames is None or filenames == [] or filenames[0] == '':
                return False

        filenames.sort(key=RUtils.natural_sort)

        # # Open file names and read annotations.
        maxid: int = 0

        if not replace:
            [allfiducials, allqpolygonfs, allpolyline_ids] = self.pjs.copy_annotations()
            allpolylines = RUtils.qpolygonfs2coordinatelists(allqpolygonfs)

            maxid = max([0] + [max(slice_ids) for slice_ids in allpolyline_ids if
                               slice_ids])  # the [0] ensures no error if there are no annotations.

        else:
            allfiducials = [[] for _ in range(self.pjs.n_frames)]
            allpolylines = [[] for _ in range(self.pjs.n_frames)]
            allpolyline_ids = [[] for _ in range(self.pjs.n_frames)]

        curfiducials: List[List[int]] = None
        curpolylines: List[List[int]] = None
        curpolyline_ids: List[List[int]] = None

        for fname in filenames:
            try:
                matlabVars = scipy.io.loadmat(fname, struct_as_record=False)
            except (IOError, OSError) as ex:
                print(ex)
                return False

            ud = matlabVars['ud'][0][0]

            if ud.rfiducials.size > 0:
                curfiducials = numpy.asarray(numpy.transpose(ud.rfiducials, [2, 0, 1]), dtype=int)
            else:
                curfiducials = [[] for _ in range(self.pjs.n_frames)]

            if ud.rpolygons.size > 0:
                curpolylines = numpy.transpose(ud.rpolygons, [2, 0, 1])
            else:
                curpolylines = [[] for _ in range(self.pjs.n_frames)]

            curpolyline_ids = [[] for _ in range(self.pjs.n_frames)]

            for slice_index, slice_polylines in enumerate(curpolylines):
                for thepolyline_index, thepolyline in enumerate(slice_polylines):
                    if thepolyline[0].size > 0:
                        curpolyline_ids[slice_index].append(thepolyline_index + 1)

            for i in range(len(allfiducials)):
                allfiducials[i].extend(curfiducials[i])

            for i in range(len(allpolylines)):
                allpolylines[i].extend([apolylinearray[0].tolist() for apolylinearray in curpolylines[i]])
                allpolyline_ids[i].extend((numpy.asarray(curpolyline_ids[i]) + maxid).tolist())

            themaxid = max([max(slice_ids) for slice_ids in curpolyline_ids if slice_ids])
            maxid = max(maxid, themaxid)

        if image_file == '':  # Create a blank image.
            # Find maxx, maxy, and maxz, and create a numpy.ndarray with those dimensions + 1.
            maxz = len(allfiducials)
            maxx = 0
            maxy = 0

            # Fiducials and polylines are stored as (x, y) coordinates.
            # numpy.ndarrays are created with (slices, rows, cols).
            for slice_fiducials in allfiducials:
                for afiducial in slice_fiducials:
                    maxx = max(maxx, afiducial[0])
                    maxy = max(maxy, afiducial[1])

            for slice_polylines in allpolylines:
                for apolyline in slice_polylines:
                    for afiducial in apolyline:
                        maxx = max(maxx, afiducial[0])
                        maxy = max(maxy, afiducial[1])

            virtual_image = numpy.zeros((maxz, int(maxy + 1), int(maxx + 1)), dtype=int)
            self.pjs.io.cbLoadArray(virtual_image)

        elif image_file is not False and image_file is not None:
            self.pjs.io.cbLoadTimeSeries(image_file)

        self.pjs.annotations.cbDeleteAllAnn()

        self.pjs.undo_stack.push(
            {'changetype': undo_modes.MULTI, 'frame': (0, self.pjs.n_frames), 'index': (True, True),
             'details': self.pjs.copy_annotations()})

        for slice_index, slice_fiducials in enumerate(allfiducials):
            for afiducial in slice_fiducials:
                if afiducial[-1] >= 0:
                    self.pjs.addFiducial(int(afiducial[0]), int(afiducial[1]), slice_index, paint=False)

        for i, theframepolylines in enumerate(allpolylines):
            if i < self.pjs.n_frames:
                for thepolyline in theframepolylines:
                    if thepolyline != [] and thepolyline != [[]]:  # the condition comparing to [[]] is for an edge case with bad user annotations in siesta.
                        if self.pjs.close_all_polylines and thepolyline[0] != thepolyline[-1]:
                            thepolyline.append(thepolyline[0])
                        self.pjs.addPolyline(thepolyline, i, paint=False)

        self.pjs.repaint()

        # Modify current path.
        self.pjs.cwd = os.path.dirname(filenames[0])

        return True


    def cbExportSIESTAAnnotations(self, filename: Optional[str] = None,
                                  message: str = "Export SIESTA annotations ...") -> bool:
        """
        Export annotations in Matlab format. These annotations can be read by both Matlab and Python.
        See https://www.mathworks.com/help/matlab/matlab_external/handling-data-returned-from-python.html for details.

        Something important here is to send float arrays to Matlab, otherwise there are errors when
        Matlab tries to conduct certain operations on the arrays.

        :param filename: name of the file used to store the annotations.
        :param message: string to display as the dialog box title.
        :return: True if the annotations are exported, False otherwise.
        """

        # Get file name.
        if filename is None or filename is False:  # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            fname = QtWidgets.QFileDialog.getSaveFileName(None, message,
                                                          self.pjs.cwd,
                                                          filter='SIESTA annotations (*.mat)')  # fname[0] is the full filename, fname[1] is the filter used.
            # If cancel ...
            filename = fname[0]
            if filename == '':
                return False
        elif filename == '':
            _, fname = os.path.split(self.pjs.filename)
            filename = os.path.join(self.pjs.cwd, fname)

        # Modify current path.
        self.pjs.cwd = os.path.dirname(filename)

        # Find max number of fiducials and polygons in any frame.
        rnfiducials = numpy.zeros(self.pjs.n_frames, dtype=float)
        max_n_polylines = 0
        for iSlice, thefiducials in enumerate(self.pjs.fiducials):
            rnfiducials[iSlice] = len(thefiducials)
            if len(self.pjs.polylines[iSlice]) > max_n_polylines:
                max_n_polylines = len(self.pjs.polylines[iSlice])

        # Create the fiducial array to send to Matlab.
        max_n_fiducials = numpy.amax(rnfiducials)
        rfiducials = -1. * numpy.ones([int(max_n_fiducials), 3, self.pjs.n_frames], dtype=float)

        # rpolylines is a list, which in Matlab is a cell.
        # The right dimensions are (max_n_fiducials, 1, self.pjs.n_frames)
        # Need to add the singlet second dimensions. The order of the square brackets is critical here.
        rpolylines = [[[[] for _ in range(self.pjs.n_frames)]] for _ in range(max_n_polylines)]

        for iSlice, thefiducials in enumerate(self.pjs.fiducials):
            if thefiducials:
                thefiducials_array = numpy.asarray(thefiducials, dtype=float)
                rfiducials[0:thefiducials_array.shape[0], 0:2, iSlice] = thefiducials_array
                rfiducials[0:thefiducials_array.shape[0], 2, iSlice] = float(iSlice)

            thepolylines = self.pjs.polylines[iSlice]

            if thepolylines:
                for iPoly, thePoly in enumerate(thepolylines):
                    theintpoly = [[float(pnt.x()), float(pnt.y())] for pnt in thePoly]

                    rpolylines[iPoly][0][iSlice] = theintpoly

        imsize = [float(self.pjs.width), float(self.pjs.height), float(self.pjs.n_frames)]

        # Build dictionary with the variables. Dictionaries are converted to structs
        ud = {
            'rfiducials': rfiducials,
            'rpolygons': rpolylines,
            'rnfiducials': rnfiducials,
            'imsize': imsize,
        }

        # Open file name and save annotations.
        try:
            scipy.io.savemat(filename, {'ud': ud}, appendmat=True)
        except (IOError, OSError) as ex:
            print(ex)
            return False

        # Modify current path.
        self.pjs.cwd = os.path.dirname(filename)

        self.pjs.statusbar.showMessage(f'Exported annotations to {filename}.')

        return True






"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional

from PyQt6 import QtCore, QtWidgets

from pyjamas.dialogs.classifierdialogABC import ClassifierDialogABC
import pyjamas.rimage.rimml.rimlr as rimlr


class LRDialog(ClassifierDialogABC):
    misclass_penalty_C: float = rimlr.lr.DEFAULT_C

    def __init__(self):
        super().__init__()

    def setupUi(self, LR, parameters: Optional[dict] = None):
        if parameters is None or parameters is False:
            parameters = {
                'positive_training_folder': LRDialog.positive_training_folder,
                'negative_training_folder': LRDialog.negative_training_folder,
                'hard_negative_training_folder': LRDialog.hard_negative_training_folder,
                'histogram_of_gradients': LRDialog.histogram_of_gradients,
                'train_image_size': LRDialog.train_image_size,
                'step_sz': LRDialog.step_sz,
                'misclass_penalty_C': LRDialog.misclass_penalty_C,
            }

        LRDialog.positive_training_folder = parameters.get('positive_training_folder', LRDialog.positive_training_folder)
        LRDialog.negative_training_folder = parameters.get('negative_training_folder', LRDialog.negative_training_folder)
        LRDialog.hard_negative_training_folder = parameters.get('hard_negative_training_folder', LRDialog.hard_negative_training_folder)
        LRDialog.histogram_of_gradients = parameters.get('histogram_of_gradients', LRDialog.histogram_of_gradients)
        LRDialog.train_image_size = parameters.get('train_image_size', LRDialog.train_image_size)
        LRDialog.step_sz = parameters.get('step_sz', LRDialog.step_sz)
        LRDialog.misclass_penalty_C = parameters.get('misclass_penalty_C', LRDialog.misclass_penalty_C)

        LR.setObjectName("LR")
        LR.resize(614, 432)
        self.buttonBox = QtWidgets.QDialogButtonBox(LR)
        self.buttonBox.setGeometry(QtCore.QRect(240, 374, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.groupBox_2 = QtWidgets.QGroupBox(LR)
        self.groupBox_2.setGeometry(QtCore.QRect(30, 26, 551, 121))
        self.groupBox_2.setObjectName("groupBox_2")
        self.label = QtWidgets.QLabel(self.groupBox_2)
        self.label.setGeometry(QtCore.QRect(31, 26, 141, 24))
        self.label.setObjectName("label")
        self.positive_training_folder_edit = QtWidgets.QLineEdit(self.groupBox_2)
        self.positive_training_folder_edit.setGeometry(QtCore.QRect(220, 30, 261, 21))
        self.positive_training_folder_edit.setObjectName("positive_training_folder_edit")
        self.positive_training_folder_edit.setText(LRDialog.positive_training_folder)
        self.btnSavePositive = QtWidgets.QToolButton(self.groupBox_2)
        self.btnSavePositive.setGeometry(QtCore.QRect(490, 30, 26, 22))
        self.btnSavePositive.setObjectName("btnSavePositive")
        self.btnSavePositive.clicked.connect(self._open_positive_folder_dialog)
        self.negative_training_folder_edit = QtWidgets.QLineEdit(self.groupBox_2)
        self.negative_training_folder_edit.setGeometry(QtCore.QRect(220, 60, 261, 21))
        self.negative_training_folder_edit.setObjectName("negative_training_folder_edit")
        self.negative_training_folder_edit.setText(LRDialog.negative_training_folder)
        self.label_2 = QtWidgets.QLabel(self.groupBox_2)
        self.label_2.setGeometry(QtCore.QRect(31, 56, 141, 24))
        self.label_2.setObjectName("label_2")
        self.btnSaveNegative = QtWidgets.QToolButton(self.groupBox_2)
        self.btnSaveNegative.setGeometry(QtCore.QRect(490, 60, 26, 22))
        self.btnSaveNegative.setObjectName("btnSaveNegative")
        self.btnSaveNegative.clicked.connect(self._open_negative_folder_dialog)
        self.label_3 = QtWidgets.QLabel(self.groupBox_2)
        self.label_3.setGeometry(QtCore.QRect(31, 86, 181, 24))
        self.label_3.setObjectName("label_3")
        self.btnSaveHard = QtWidgets.QToolButton(self.groupBox_2)
        self.btnSaveHard.setGeometry(QtCore.QRect(490, 90, 26, 22))
        self.btnSaveHard.setObjectName("btnSaveHard")
        self.btnSaveHard.clicked.connect(self._open_hard_folder_dialog)
        self.hard_negative_training_folder_edit = QtWidgets.QLineEdit(self.groupBox_2)
        self.hard_negative_training_folder_edit.setGeometry(QtCore.QRect(220, 90, 261, 21))
        self.hard_negative_training_folder_edit.setObjectName("hard_negative_training_folder_edit")
        self.hard_negative_training_folder_edit.setText(LRDialog.hard_negative_training_folder)
        self.groupBox_3 = QtWidgets.QGroupBox(LR)
        self.groupBox_3.setGeometry(QtCore.QRect(30, 156, 251, 61))
        self.groupBox_3.setObjectName("groupBox_3")
        self.label_4 = QtWidgets.QLabel(self.groupBox_3)
        self.label_4.setGeometry(QtCore.QRect(31, 28, 141, 24))
        self.label_4.setObjectName("label_4")
        self.lnWidth = QtWidgets.QLineEdit(self.groupBox_3)
        self.lnWidth.setGeometry(QtCore.QRect(70, 30, 31, 21))
        self.lnWidth.setObjectName("lnWidth")
        self.lnWidth.setText(str(LRDialog.train_image_size[1]))
        self.lnHeight = QtWidgets.QLineEdit(self.groupBox_3)
        self.lnHeight.setGeometry(QtCore.QRect(170, 30, 31, 21))
        self.lnHeight.setObjectName("lnHeight")
        self.lnHeight.setText(str(LRDialog.train_image_size[0]))
        self.label_5 = QtWidgets.QLabel(self.groupBox_3)
        self.label_5.setGeometry(QtCore.QRect(120, 28, 141, 24))
        self.label_5.setObjectName("label_5")
        self.label_5.raise_()
        self.label_4.raise_()
        self.lnWidth.raise_()
        self.lnHeight.raise_()
        self.groupBox_4 = QtWidgets.QGroupBox(LR)
        self.groupBox_4.setGeometry(QtCore.QRect(30, 226, 551, 61))
        self.groupBox_4.setObjectName("groupBox_4")
        self.checkBoxHOG = QtWidgets.QCheckBox(self.groupBox_4)
        self.checkBoxHOG.setGeometry(QtCore.QRect(30, 30, 161, 20))
        self.checkBoxHOG.setObjectName("checkBoxHOG")
        self.checkBoxHOG.setChecked(LRDialog.histogram_of_gradients)
        self.groupBox_5 = QtWidgets.QGroupBox(LR)
        self.groupBox_5.setGeometry(QtCore.QRect(30, 296, 551, 61))
        self.groupBox_5.setObjectName("groupBox_5")
        self.lnC = QtWidgets.QLineEdit(self.groupBox_5)
        self.lnC.setGeometry(QtCore.QRect(210, 30, 31, 21))
        self.lnC.setObjectName("lnC")
        self.lnC.setText(str(LRDialog.misclass_penalty_C))
        self.label_8 = QtWidgets.QLabel(self.groupBox_5)
        self.label_8.setGeometry(QtCore.QRect(30, 28, 181, 24))
        self.label_8.setObjectName("label_8")
        self.label_8.raise_()
        self.lnC.raise_()
        self.groupBox_6 = QtWidgets.QGroupBox(LR)
        self.groupBox_6.setGeometry(QtCore.QRect(300, 155, 281, 61))
        self.groupBox_6.setObjectName("groupBox_6")
        self.label_12 = QtWidgets.QLabel(self.groupBox_6)
        self.label_12.setGeometry(QtCore.QRect(31, 28, 141, 24))
        self.label_12.setObjectName("label_12")
        self.lnRow = QtWidgets.QLineEdit(self.groupBox_6)
        self.lnRow.setGeometry(QtCore.QRect(70, 30, 31, 21))
        self.lnRow.setObjectName("lnRow")
        self.lnRow.setText(str(LRDialog.step_sz[0]))
        self.lnColumn = QtWidgets.QLineEdit(self.groupBox_6)
        self.lnColumn.setGeometry(QtCore.QRect(180, 30, 31, 21))
        self.lnColumn.setObjectName("lnColumn")
        self.lnColumn.setText(str(LRDialog.step_sz[1]))
        self.label_13 = QtWidgets.QLabel(self.groupBox_6)
        self.label_13.setGeometry(QtCore.QRect(120, 28, 141, 24))
        self.label_13.setObjectName("label_13")
        self.label_13.raise_()
        self.label_12.raise_()
        self.lnRow.raise_()
        self.lnColumn.raise_()

        self.retranslateUi(LR)
        self.buttonBox.accepted.connect(LR.accept)
        self.buttonBox.rejected.connect(LR.reject)
        QtCore.QMetaObject.connectSlotsByName(LR)

    def retranslateUi(self, LR):
        _translate = QtCore.QCoreApplication.translate
        LR.setWindowTitle(_translate("LR", "Train logistic regression model"))
        self.groupBox_2.setTitle(_translate("LR", "Project files"))
        self.label.setText(_translate("LR", "positive training folder"))
        self.btnSavePositive.setText(_translate("LR", "..."))
        self.label_2.setText(_translate("LR", "negative training folder"))
        self.btnSaveNegative.setText(_translate("LR", "..."))
        self.label_3.setText(_translate("LR", "hard negative training folder"))
        self.btnSaveHard.setText(_translate("LR", "..."))
        self.groupBox_3.setTitle(_translate("LR", "Training image size"))
        self.label_4.setText(_translate("LR", "width"))
        self.label_5.setText(_translate("LR", "height"))
        self.groupBox_4.setTitle(_translate("LR", "Image features"))
        self.checkBoxHOG.setText(_translate("LR", "histogram of gradients"))
        self.groupBox_5.setTitle(_translate("LR", "Logistic Regression parameters"))
        self.label_8.setText(_translate("LR", "C (misclassification penalty)"))
        self.groupBox_6.setTitle(_translate("LR", "Image step size"))
        self.label_12.setText(_translate("LR", "rows"))
        self.label_13.setText(_translate("LR", "columns"))

    def parameters(self) -> dict:
        LRDialog.positive_training_folder = self.positive_training_folder_edit.text()
        LRDialog.negative_training_folder = self.negative_training_folder_edit.text()
        LRDialog.hard_negative_training_folder = self.hard_negative_training_folder_edit.text()
        LRDialog.histogram_of_gradients = self.checkBoxHOG.isChecked()
        LRDialog.train_image_size = int(self.lnHeight.text()), int(self.lnWidth.text())
        LRDialog.misclass_penalty_C = float(self.lnC.text())
        LRDialog.step_sz = (int(self.lnRow.text()), int(self.lnColumn.text()))

        return {
            'positive_training_folder': LRDialog.positive_training_folder,
            'negative_training_folder': LRDialog.negative_training_folder,
            'hard_negative_training_folder': LRDialog.hard_negative_training_folder,
            'histogram_of_gradients': LRDialog.histogram_of_gradients,
            'train_image_size': LRDialog.train_image_size,
            'step_sz': LRDialog.step_sz,
            'misclass_penalty_C': LRDialog.misclass_penalty_C,
        }



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from matplotlib.figure import Figure
from matplotlib.axes import Axes
import numpy
from PyQt6 import QtCore, QtWidgets

from pyjamas.dialogs.matplotlibdialog import MatplotlibDialog
from pyjamas.pjscore import PyJAMAS
from pyjamas.rimage.rimutils import rimutils


class AdjustContrastDialog(object):
    min_pix_percentile: int = 0
    max_pix_percentile: int = 100

    def __init__(self, pjs: PyJAMAS):
        super().__init__()

        self.pjs = pjs
        self.histogram_window: MatplotlibDialog = None

        self.prev_min_pix_percentile: int = self.pjs.min_pix_percentile
        self.prev_max_pix_percentile: int = self.pjs.max_pix_percentile

    def setupUi(self, Dialog, min_percentile=None, max_percentile=None):

        if min_percentile is None or min_percentile is False:
            min_percentile = AdjustContrastDialog.min_pix_percentile

        if max_percentile is None or max_percentile is False:
            max_percentile = AdjustContrastDialog.max_pix_percentile

        self.dialog = Dialog
        self.dialog.setObjectName("Dialog")
        self.dialog.resize(221, 245)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.dialog.sizePolicy().hasHeightForWidth())
        self.dialog.setSizePolicy(sizePolicy)
        self.buttonBox = QtWidgets.QDialogButtonBox(self.dialog)
        self.buttonBox.setGeometry(QtCore.QRect(-150, 200, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel | QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.groupBox_3 = QtWidgets.QGroupBox(self.dialog)
        self.groupBox_3.setGeometry(QtCore.QRect(20, 110, 181, 91))
        self.groupBox_3.setObjectName("groupBox_3")
        self.sbmin = QtWidgets.QSpinBox(self.groupBox_3)
        self.sbmin.setGeometry(QtCore.QRect(5, 60, 48, 24))
        self.sbmin.setObjectName("sbmin")
        self.sbmin.setMinimum(0)
        self.sbmin.setMaximum(100)
        self.sbmin.setValue(min_percentile)
        self.sbmax = QtWidgets.QSpinBox(self.groupBox_3)
        self.sbmax.setGeometry(QtCore.QRect(130, 60, 48, 24))
        self.sbmax.setObjectName("sbmax")
        self.sbmax.setMinimum(0)
        self.sbmax.setMaximum(100)
        self.sbmax.setValue(max_percentile)
        self.pushButton = QtWidgets.QPushButton(self.groupBox_3)
        self.pushButton.setGeometry(QtCore.QRect(55, 57, 71, 32))
        self.pushButton.setObjectName("pushButton")
        self.minSlider = QtWidgets.QSlider(self.groupBox_3)
        self.minSlider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.minSlider.setGeometry(QtCore.QRect(4, 20, 171, 22))
        self.minSlider.setObjectName("minSlider")
        self.minSlider.setRange(0, 100)
        self.minSlider.setValue(min_percentile)
        self.minSlider.setStyleSheet("""
                                        QSlider::groove:horizontal {
                                        border: 1px solid #bbb;
                                        background: white;
                                        height: 10px;
                                        border-radius: 4px;
                                        }

                                        QSlider::sub-page:horizontal {
                                        background: qlineargradient(x1: 0, y1: 0,    x2: 0, y2: 1,
                                            stop: 0 #66e, stop: 1 #bbf);
                                        background: qlineargradient(x1: 0, y1: 0.2, x2: 1, y2: 1,
                                            stop: 0 #bbf, stop: 1 #55f);
                                        border: 1px solid #777;
                                        height: 10px;
                                        border-radius: 4px;
                                        }

                                        QSlider::add-page:horizontal {
                                        background: #fff;
                                        border: 1px solid #777;
                                        height: 10px;
                                        border-radius: 4px;
                                        }

                                        QSlider::handle:horizontal {
                                        background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                                            stop:0 #eee, stop:1 #ccc);
                                        border: 1px solid #777;
                                        width: 13px;
                                        margin-top: -2px;
                                        margin-bottom: -2px;
                                        border-radius: 4px;
                                        }

                                        QSlider::handle:horizontal:hover {
                                        background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                                            stop:0 #fff, stop:1 #ddd);
                                        border: 1px solid #444;
                                        border-radius: 4px;
                                        }

                                        QSlider::sub-page:horizontal:disabled {
                                        background: #bbb;
                                        border-color: #999;
                                        }

                                        QSlider::add-page:horizontal:disabled {
                                        background: #eee;
                                        border-color: #999;
                                        }

                                        QSlider::handle:horizontal:disabled {
                                        background: #eee;
                                        border: 1px solid #aaa;
                                        border-radius: 4px;
                                        }
                                    """)

        self.maxSlider = QtWidgets.QSlider(self.groupBox_3)
        self.maxSlider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.maxSlider.setGeometry(QtCore.QRect(4, 40, 171, 22))
        self.maxSlider.setObjectName("maxSlider")
        self.maxSlider.setRange(0, 100)
        self.maxSlider.setValue(max_percentile)
        self.maxSlider.setStyleSheet("""
                                        QSlider::groove:horizontal {
                                        border: 1px solid #bbb;
                                        background: white;
                                        height: 10px;
                                        border-radius: 4px;
                                        }
                                        
                                        QSlider::sub-page:horizontal {
                                        background: qlineargradient(x1: 0, y1: 0,    x2: 0, y2: 1,
                                            stop: 0 #e66, stop: 1 #fbb);
                                        background: qlineargradient(x1: 0, y1: 0.2, x2: 1, y2: 1,
                                            stop: 0 #fbb, stop: 1 #f55);
                                        border: 1px solid #777;
                                        height: 10px;
                                        border-radius: 4px;
                                        }
                                        
                                        QSlider::add-page:horizontal {
                                        background: #fff;
                                        border: 1px solid #777;
                                        height: 10px;
                                        border-radius: 4px;
                                        }
                                        
                                        QSlider::handle:horizontal {
                                        background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                                            stop:0 #eee, stop:1 #ccc);
                                        border: 1px solid #777;
                                        width: 13px;
                                        margin-top: -2px;
                                        margin-bottom: -2px;
                                        border-radius: 4px;
                                        }
                                        
                                        QSlider::handle:horizontal:hover {
                                        background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                                            stop:0 #fff, stop:1 #ddd);
                                        border: 1px solid #444;
                                        border-radius: 4px;
                                        }
                                        
                                        QSlider::sub-page:horizontal:disabled {
                                        background: #bbb;
                                        border-color: #999;
                                        }
                                        
                                        QSlider::add-page:horizontal:disabled {
                                        background: #eee;
                                        border-color: #999;
                                        }
                                        
                                        QSlider::handle:horizontal:disabled {
                                        background: #eee;
                                        border: 1px solid #aaa;
                                        border-radius: 4px;
                                        }
                                    """)

        self.histogram_window = MatplotlibDialog(self.dialog, self.generate_figure(), '', False)
        self.histogram_window.setGeometry(QtCore.QRect(10, 10, 201, 91))
        self.histogram_window.setObjectName("histogram")

        self.retranslateUi()
        self.pushButton.pressed.connect(self._auto_adjust)
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)
        self.sbmin.valueChanged.connect(self._update_min_sb)
        self.sbmax.valueChanged.connect(self._update_max_sb)
        self.minSlider.valueChanged.connect(self._update_min_slider)
        self.maxSlider.valueChanged.connect(self._update_max_slider)
        QtCore.QMetaObject.connectSlotsByName(self.dialog)

    def accept(self):
        AdjustContrastDialog.min_pix_percentile = self.sbmin.value()
        AdjustContrastDialog.max_pix_percentile = self.sbmax.value()

        self.dialog.accept()

    def reject(self):
        self.pjs.min_pix_percentile = self.prev_min_pix_percentile
        self.pjs.max_pix_percentile = self.prev_max_pix_percentile

        self.pjs.displayData()

        self.dialog.reject()

    def retranslateUi(self):
        _translate = QtCore.QCoreApplication.translate
        self.dialog.setWindowTitle(_translate("Dialog", "Adjust contrast"))
        self.groupBox_3.setTitle(_translate("Dialog", "Min and max percentile"))
        self.pushButton.setText(_translate("Dialog", "Auto"))

    def _update_min_sb(self) -> bool:
        self.minSlider.valueChanged.disconnect()
        self.maxSlider.valueChanged.disconnect()
        self.sbmin.valueChanged.disconnect()
        self.sbmax.valueChanged.disconnect()

        v = self.sbmin.value()

        if self.sbmax.value() < v:
            self.sbmax.setValue(v)
            self.maxSlider.setValue(v)
        self.minSlider.setValue(v)

        self.minSlider.valueChanged.connect(self._update_min_slider)
        self.maxSlider.valueChanged.connect(self._update_max_slider)
        self.sbmin.valueChanged.connect(self._update_min_sb)
        self.sbmax.valueChanged.connect(self._update_max_sb)

        self.update_curslice()
        self._update_histogram()
        self.dialog.raise_()
        self.dialog.activateWindow()

        return True

    def _update_max_sb(self) -> bool:
        self.minSlider.valueChanged.disconnect()
        self.maxSlider.valueChanged.disconnect()
        self.sbmin.valueChanged.disconnect()
        self.sbmax.valueChanged.disconnect()

        v = self.sbmax.value()

        if self.sbmin.value() > v:
            self.sbmin.setValue(v)
            self.minSlider.setValue(v)
        self.maxSlider.setValue(v)

        self.minSlider.valueChanged.connect(self._update_max_slider)
        self.maxSlider.valueChanged.connect(self._update_max_slider)
        self.sbmin.valueChanged.connect(self._update_min_sb)
        self.sbmax.valueChanged.connect(self._update_max_sb)

        self.update_curslice()
        self._update_histogram()
        self.dialog.raise_()
        self.dialog.activateWindow()

        return True

    def _update_min_slider(self) -> bool:
        self.minSlider.valueChanged.disconnect()
        self.maxSlider.valueChanged.disconnect()
        self.sbmin.valueChanged.disconnect()
        self.sbmax.valueChanged.disconnect()

        v = self.minSlider.value()

        if self.maxSlider.value() >= v:
            self.sbmin.setValue(v)

        else:
            self.maxSlider.setValue(v)
            self.sbmin.setValue(v)
            self.sbmax.setValue(v)

        self.minSlider.valueChanged.connect(self._update_min_slider)
        self.maxSlider.valueChanged.connect(self._update_max_slider)
        self.sbmin.valueChanged.connect(self._update_min_sb)
        self.sbmax.valueChanged.connect(self._update_max_sb)

        self.update_curslice()
        self._update_histogram()
        self.dialog.raise_()
        self.dialog.activateWindow()

        return True

    def _update_max_slider(self) -> bool:
        self.minSlider.valueChanged.disconnect()
        self.maxSlider.valueChanged.disconnect()
        self.sbmin.valueChanged.disconnect()
        self.sbmax.valueChanged.disconnect()

        v = self.maxSlider.value()

        if self.minSlider.value() <= v:
            self.sbmax.setValue(v)

        else:
            self.minSlider.setValue(v)
            self.sbmax.setValue(v)
            self.sbmin.setValue(v)

        self.minSlider.valueChanged.connect(self._update_min_slider)
        self.maxSlider.valueChanged.connect(self._update_max_slider)
        self.sbmin.valueChanged.connect(self._update_min_sb)
        self.sbmax.valueChanged.connect(self._update_max_sb)

        self.update_curslice()
        self._update_histogram()
        self.dialog.raise_()
        self.dialog.activateWindow()

        return True

    def _auto_adjust(self) -> bool:
        self.sbmin.setValue(50)
        self.sbmax.setValue(99)

        return True

    def parameters(self) -> dict:
        # AdjustContrastDialog.min_pix_percentile = self.sbmin.value()
        # AdjustContrastDialog.max_pix_percentile = self.sbmax.value()

        return {'min_percentile': self.sbmin.value(),
                         'max_percentile': self.sbmax.value()}

    def update_curslice(self) -> bool:
        parameters = self.parameters()

        self.pjs.min_pix_percentile = parameters.get('min_percentile', 0)
        self.pjs.max_pix_percentile = parameters.get('max_percentile', 100)

        self.pjs.displayData()

        return True

    def _update_histogram(self) -> bool:
        if self.line_min is not None and self.line_max is not None:
            self.line_min.pop(0).remove()
            self.line_max.pop(0).remove()

            ax = self.histogram_window.figure.gca()

            self._paint_lines(ax)

            self.histogram_window.canvas.draw()

        return True

    def generate_figure(self) -> Figure:
        figure = rimutils.histogram_figure(self.pjs.imagedata, False)
        ax = figure.gca()

        ylimits = ax.get_ylim()

        self._paint_lines(ax)

        #ax.set_xlim((0, self.pjs.imagedata.max()))

        ax.set_ylim(ylimits)

        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)

        return figure

    def _paint_lines(self, ax: Axes) -> bool:
        min_val = numpy.percentile(self.pjs.imagedata, self.pjs.min_pix_percentile)
        max_val = numpy.percentile(self.pjs.imagedata, self.pjs.max_pix_percentile)

        self.line_min = ax.plot((min_val, min_val), (0, 1), 'b')
        self.line_max = ax.plot((max_val, max_val), (0, 1), 'r')

        return True


"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Tuple

from PyQt6 import QtCore, QtWidgets
from pyjamas.rimage.rimml.classifier_types import classifier_types


class ClassifierTypeDialog(object):
    # These are class variables so that, when their values change, they change, and next time the same dialog is
    # open, the most recent values are preserved.
    classifier_type = classifier_types.UNKNOWN.value

    def __init__(self):
        super().__init__()

    def setupUi(self, Dialog, classifier_type: str = classifier_types.UNKNOWN.value):

        classifier_identities = set(thetype.value for thetype in classifier_types)
        if classifier_type not in classifier_identities:
            self.classifier_type = classifier_types.UNKNOWN.value
        else:
            self.classifier_type = classifier_type

        Dialog.setObjectName("Dialog")
        Dialog.resize(210, 100)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Dialog.sizePolicy().hasHeightForWidth())
        Dialog.setSizePolicy(sizePolicy)

        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(-160, 60, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")

        self.comboBox = QtWidgets.QComboBox(Dialog)
        self.comboBox.setGeometry(QtCore.QRect(20, 20, 170, 30))
        self.comboBox.addItems(classifier_identities)
        self.comboBox.setCurrentText(self.classifier_type)
        self.comboBox.setObjectName("combobox")

        self.retranslateUi(Dialog)
        self.buttonBox.rejected.connect(Dialog.reject)
        self.buttonBox.accepted.connect(Dialog.accept)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Select classifier type"))

    def parameters(self) -> Tuple[int, int]:
        ClassifierTypeDialog.classifier_type = self.comboBox.currentText()

        return ClassifierTypeDialog.classifier_type



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional

from sklearn.svm import SVC

from pyjamas.rimage.rimml.rimclassifier import rimclassifier
from pyjamas.rimage.rimml.classifier_types import classifier_types


class svm(rimclassifier):
    KERNEL_TYPES = ('linear', 'rbf')
    CLASSIFIER_TYPE: str = classifier_types.SUPPORT_VECTOR_MACHINE.value

    DEFAULT_KERNEL_TYPE: int = 1

    # Missed class penalty: small values (0.05) result in some additional nuclei detected and others
    # not. Large values (100) do not seem to change the result with respect to 1.0.
    DEFAULT_C: float = 1.0

    def __init__(self, parameters: Optional[dict] = None):
        super().__init__(parameters)

        # SVM-specific parameters.
        misclass_penalty_C: float = parameters.get('C', svm.DEFAULT_C)
        kernel_type: int = parameters.get('kernel_type', svm.DEFAULT_KERNEL_TYPE)
        self.classifier = parameters.get('classifier', SVC(kernel=svm.KERNEL_TYPES[kernel_type],
                                                           gamma='scale',
                                                           C=misclass_penalty_C,
                                                           random_state=rimclassifier.DEFAULT_SEED,
                                                           probability=True))



from enum import Enum


class classifier_types(Enum):
    # Classifier types.
    UNKNOWN: str = "Unknown"
    SUPPORT_VECTOR_MACHINE: str = "Support vector machine"
    LOGISTIC_REGRESSION: str = "Logistic regression"
    UNET: str = "U-Net"
    RESCUNET: str = "ReSCU-Net"



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import List

import numba

numba.warnings.filterwarnings('ignore', '', numba.NumbaWarning)

import numpy
import scipy.ndimage as sciim
import PyQt6.QtGui
from shapely.geometry import LineString

from pyjamas.rannotations import *
from pyjamas.rutils import RUtils


class RPolyline(rannotation.RAnnotation):
    '''
    Represents a multisegment line (open) or polygon (closed) by providing a list of points (list, tuple, ndarray, etc),
    and a flag to indicate whether the curve is closed (1) or open (1).
    '''

    # Define static variables here.

    def __init__(self, point_list: object):
        """
        Creates a polyline (closed_flag = 1) or a trajectory (closed_flag = 0).
        point_list is a list or tuple that will be converted into an ndarray. It can have matching first and last
          points if it is a polyline.

        :param point_list: list, tuple, ndarray, ... Internally, a numpy.ndarray.
        :param closed_flag: bool
        """

        super().__init__()

        # Define object variables here.
        if type(point_list) == PyQt6.QtGui.QPolygonF:
            thepoints = [[apoint.x(), apoint.y()] for apoint in point_list]
            self.points = numpy.asarray(thepoints)
            self.closed_flag = point_list.isClosed()

        elif type(point_list) == list:
            self.points = numpy.asarray(point_list)  # Points that define the polyline.
            self.closed_flag = point_list[0] == point_list[-1]

        self._area = -1  # Calculate automatically as the polygon is created?
        self._perimeter = -1  # Calculate automatically as the polygon is created?

    def area(self) -> numpy.double:
        """
        Calculate the area of a planar, not self intersecting polygon:
        http://mathworld.wolfram.com/PolygonArea.html.
        Note that the area of a convex polygon is defined to be positive if the points are arranged in a
        counterclockwise order, and negative if they are in clockwise order (Beyer 1987).
        Beyer, W. H. (Ed.). CRC Standard Mathematical Tables, 28th ed. Boca Raton, FL: CRC Press, pp. 123-124, 1987.

        :return: double
        """
        if self.closed_flag:
            n = len(self.points)

            area = 0.0

            for i in range(n - 1):
                area += self.points[i, 0] * self.points[i + 1, 1] - self.points[i + 1, 0] * self.points[i, 1]

            area += self.points[n - 1, 0] * self.points[0, 1] - self.points[0, 0] * self.points[n - 1, 1]

            self._area = abs(area / 2)

        else:
            self._area = 0.0

        return self._area

    def perimeter(self) -> numpy.double:
        """
        Calculate the perimeter of a polygon.

        :return: double
        """

        shifted_points = numpy.roll(self.points, 2)
        diff_points = self.points - shifted_points

        if not self.closed_flag:
            diff_points = diff_points[1:]

        self._perimeter = numpy.sum(numpy.linalg.norm(diff_points, 2, axis=1))

        return self._perimeter

    def pixel_values(self, image: numpy.ndarray, line_width: numpy.uint8 = 3) -> \
            (numpy.double, numpy.double, numpy.double, numpy.double):
        """
        Calculate the mean and standard deviation of the pixel values under a polygon and inside the same polygon,
        and return as a tuple (perimeter_mean, interior_mean, perimeter_std, interior_std).

        :param image: numpy.ndarray
        :param line_width: brush size used to calculate boundary and interior, numpy.uint8 = 3
        :return: (numpy.double, numpy.double, numpy.double, numpy.double)
        """
        perimeter_mean = numpy.nan
        perimeter_std = numpy.nan
        interior_mean = numpy.nan
        interior_std = numpy.nan

        # Find points under the annotation and calculate their mean intensity.
        edge_mask, inside_mask = self.tomask(image.shape,
                                             line_width)  # RPolyline.bresenham_wide_line2d(self.points, line_width)
        if edge_mask.size > 0:
            perimeter_mean = numpy.mean(
                image[edge_mask[::, 0], edge_mask[::, 1]])  # image is indexed like this: image[(rows), (cols)]
            perimeter_std = numpy.std(image[edge_mask[::, 0], edge_mask[::, 1]])
        if inside_mask.size > 0:
            interior_mean = numpy.mean(image[inside_mask[::, 0], inside_mask[::, 1]])
            interior_std = numpy.std(image[inside_mask[::, 0], inside_mask[::, 1]])

        return perimeter_mean, interior_mean, perimeter_std, interior_std

    @staticmethod
    @numba.jit  # makes function 6x faster
    def bresenham_singlepix_line3d(p1: numpy.ndarray, p2: numpy.ndarray, precision: numpy.double = 0.) -> (
            numpy.ndarray, numpy.ndarray, numpy.ndarray):
        """"
        Generate X Y Z coordinates of a 3D Bresenham's line between
        two given points. The line has a width of one pixel.

              Usage: [X Y Z] = bresenham_line3d(p1, p2, precision)

              p1	- vector for Point1, where p1 = [x1 y1 z1]

              p2	- vector for Point2, where p2 = [x2 y2 z2]

              precision (optional) - Although according to Bresenham's line
                algorithm, point coordinates x1 y1 z1 and x2 y2 z2 should
                be integer numbers, this program extends its limit to all
                real numbers. If any of them are floating numbers, you
                should specify how many digits of decimal that you would
                like to preserve. Be aware that the length of output X Y
                Z coordinates will increase in 10 times for each decimal
                digit that you want to preserve. By default, the precision
                is 0, which means that they will be rounded to the nearest
                integer.

              X	- a set of x coordinates on Bresenham's line

              Y	- a set of y coordinates on Bresenham's line

              Z	- a set of z coordinates on Bresenham's line

              Therefore, all points in XYZ set (i.e. P(i) = [X(i) Y(i) Z(i)])
              will constitute the Bresenham's line between p1 and p2.

              Example:
                p1 = [12 37 6]     p2 = [46 3 35]
                [X Y Z] = bresenham_line3d(p1, p2)
                figure plot3(X,Y,Z,'s','markerface','b')

              This program is ported to MATLAB from:

              B.Pendleton.  line3d - 3D Bresenham's (a 3D line drawing algorithm)
              ftp://ftp.isc.org/pub/usenet/comp.sources.unix/volume26/line3d, 1992

              Which is also referenced by:

              Fischer, J., A. del Rio (2004).  A Fast Method for Applying Rigid
              Transformations to Volume Data, WSCG2004 Conference.
              http://wscg.zcu.cz/wscg2004/Papers_2004_Short/M19.pdf

              - Jimmy Shen (jimmy@rotman-baycrest.on.ca)


        :param p1: numpy.ndarray (x1, y1, z1)
        :param p2: numpy.ndarray (x2, y2, z2)
        :param precision: numpy.double=0. Although according to Bresenham's line
                algorithm, point coordinates x1 y1 z1 and x2 y2 z2 should
                be integer numbers, this program extends its limit to all
                real numbers. If any of them are floating numbers, you
                should specify how many digits of decimal that you would
                like to preserve. Be aware that the length of output X Y
                Z coordinates will increase in 10 times for each decimal
                digit that you want to preserve. By default, the precision
                is 0, which means that they will be rounded to the nearest
                integer.
        :return: (numpy.ndarray, numpy.ndarray. numpy.ndarray), X, Y, and Z coordinates, respectively, of the points
                in the Bresenham line.

        """

        if numpy.round(precision) == 0:
            precision = 0.
            p1 = numpy.round(p1)
            p2 = numpy.round(p2)

        else:
            precision = numpy.round(precision)
            p1 = numpy.round(p1 * (10 ^ precision))
            p2 = numpy.round(p2 * (10 ^ precision))

        d = numpy.int16(numpy.max(numpy.abs(p2 - p1) + 1))
        X = numpy.zeros(d)
        Y = numpy.zeros(d)
        Z = numpy.zeros(d)

        x1 = p1[0]
        y1 = p1[1]
        z1 = p1[2]

        x2 = p2[0]
        y2 = p2[1]
        z2 = p2[2]

        dx = x2 - x1
        dy = y2 - y1
        dz = z2 - z1

        ax = numpy.abs(dx) * 2
        ay = numpy.abs(dy) * 2
        az = numpy.abs(dz) * 2

        sx = numpy.sign(dx)
        sy = numpy.sign(dy)
        sz = numpy.sign(dz)

        x = x1
        y = y1
        z = z1
        idx = 0

        if ax >= numpy.maximum(ay, az):  # x dominant
            yd = ay - ax / 2
            zd = az - ax / 2

            while True:
                X[idx] = x
                Y[idx] = y
                Z[idx] = z
                idx += 1

                if x == x2:
                    break

                if yd >= 0:  # move along y
                    y += sy
                    yd -= ax

                if zd >= 0:  # move along z
                    z += sz
                    zd -= ax

                x += sx  # move along x
                yd += ay
                zd += az

        elif ay >= numpy.maximum(ax, az):  # y dominant
            xd = ax - ay / 2
            zd = az - ay / 2

            while True:
                X[idx] = x
                Y[idx] = y
                Z[idx] = z
                idx += 1

                if y == y2:
                    break

                if xd >= 0:  # move along x
                    x += sx
                    xd -= ay

                if zd >= 0:  # move along z
                    z += sz
                    zd -= ay

                y += sy  # move along y
                xd += ax
                zd += az

        elif az >= numpy.maximum(ax, ay):  # z dominant
            xd = ax - az / 2
            yd = ay - az / 2

            while True:
                X[idx] = x
                Y[idx] = y
                Z[idx] = z
                idx += 1

                if z == z2:
                    break

                if xd >= 0:  # move along x
                    x += sx
                    xd -= az

                if yd >= 0:  # move along y
                    y += sy
                    yd -= az

                z += sz  # move along z
                xd += ax
                yd += ay

        if precision != 0:
            X = X / numpy.power(10, precision)
            Y = Y / numpy.power(10, precision)
            Z = Z / numpy.power(10, precision)

        return X, Y, Z

    @classmethod
    def bresenham_wide_line2d(cls, p1: numpy.ndarray, p2: numpy.ndarray, width: numpy.uint = 0) -> (
            numpy.ndarray, numpy.ndarray):
        """
        Generate (x, y) coordinates of a 2D Bresenham line of arbitrary width.

        :param p1: (x, y) coordinates of source point.
        :param p2: (x, y) coordinates of destination point.
        :param width: line width (in pixels).
        :return: (numpy.ndarray, numpy.ndarray), first output array are x coordinates, second are y coordinates.
        """

        x1 = p1[0]
        x2 = p2[0]
        y1 = p1[1]
        y2 = p2[1]

        hw = int(numpy.floor(width / 2.))

        dx = numpy.abs(x2 - x1)
        dy = numpy.abs(y2 - y1)

        x = numpy.asarray([])
        y = numpy.asarray([])

        for ii in range(-hw, hw + 1):
            if dx > dy:
                y1new = y1 + ii
                y2new = y2 + ii
                tmpx, tmpy, _ = numpy.int16(cls.bresenham_singlepix_line3d(numpy.asarray([x1, y1new, 0]),
                                                                           numpy.asarray([x2, y2new, 0])))
            else:
                x1new = x1 + ii
                x2new = x2 + ii
                tmpx, tmpy, _ = numpy.int16(cls.bresenham_singlepix_line3d(numpy.asarray([x1new, y1, 0]),
                                                                           numpy.asarray([x2new, y2, 0])))

            x = numpy.append(x, tmpx)
            y = numpy.append(y, tmpy)
        return x, y

    def tomask(self, imshape: numpy.ndarray, brushsz: numpy.uint8) -> (numpy.ndarray, numpy.ndarray):
        '''
        Convert polygon to a mask and return the coordinates of the pixels under the polygon and inside the polygon.

        Note that numpy.ndarrays are indexed (and shaped) by (row, column), starting at (0,0).
        However, PyJAMAS stores point coordinates as (x, y) (equivalent to (col, row)) starting at (0, 0).
        :param imshape: (rows, cols)
        :param brushsz: line width.
        :return: numpy.ndarray (x, y), numpy.ndarray (x, y) coordinates of perimeter pixels and interior pixels.
        '''
        binary_image = numpy.zeros(imshape, dtype=bool)

        x = numpy.asarray([])
        y = numpy.asarray([])

        # Find the coordinates of the points under the polyline.
        for ii in range(len(self.points) - 1):
            thex, they = self.bresenham_wide_line2d(self.points[ii], self.points[ii + 1], brushsz)

            # Avoid repeating the start points by not including the first point of each segments.
            x = numpy.append(x, thex)
            y = numpy.append(y, they)

        # Get rid of any pixels not on the image.
        ind_good = numpy.nonzero((x >= 0) & (x < imshape[1]) & (y >= 0) & (y < imshape[0]))
        x = x[ind_good]
        y = y[ind_good]

        # Those points form an edge mask (we remove repeated points) after inverting coordinates
        # to go from (x, y) to (row, col).
        themask_perimeter = numpy.unique((y, x), axis=1)

        # Set all the pixels under the polyline to True in the mask.
        perimeter_pixel_indices = [(int(x), int(y)) for x, y in
                                   zip(themask_perimeter[0], themask_perimeter[1])]
        perimeter_pixel_indices = numpy.asarray(perimeter_pixel_indices)
        binary_image[perimeter_pixel_indices[:, 0], perimeter_pixel_indices[:, 1]] = True

        # The internal mask is the difference between a solid fill of the polyline and the polyline.
        binary_image2 = sciim.binary_fill_holes(binary_image) ^ binary_image
        themask_inside = binary_image2.nonzero()

        inside_pixel_indices = [(int(x), int(y)) for x, y in zip(themask_inside[0], themask_inside[1])]
        inside_pixel_indices = numpy.asarray(inside_pixel_indices)

        return perimeter_pixel_indices, inside_pixel_indices

    def control_points(self, npoints: int) -> numpy.ndarray:
        """
        Determines a set of points approximately equispaced along the polyline.

        :param npoints: how many points to identify.
        :return: numpy.ndarray containing indices of control points (for the actual coordinates, self.points[control_points_indices]).
        """

        # Create a line string and interpolate coordinates of npoints equispaced points.
        line: LineString = LineString(self.points.tolist())
        distances: numpy.ndarray = numpy.linspace(0, line.length, npoints)
        thepoints: List = [line.interpolate(distance) for distance in distances]

        # Convert point coordinates to array.
        thepoint_array: numpy.ndarray = numpy.asarray([[apoint.x, apoint.y] for apoint in thepoints])

        # Return the indices of the closest points in the polyline.
        return self.closest_points(thepoint_array)

    def closest_points(self, coordinates: numpy.ndarray) -> numpy.ndarray:
        """
        Return indices of points in the polyline closest to a certain set of coordinates.

        :param coordinates: numpy.ndarray containing the x, y coordinates of points to match.
        :return: numpy.ndarray containing the indices of the points in the polyline closest to the points in coordinates.
        """
        dist_matrix = RUtils.point2point_distances(coordinates, self.points)

        return numpy.argmin(dist_matrix, axis=1)



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional, Tuple

import matplotlib.pyplot as plt

plt.ioff()  # to keep figure windows from popping up unless show is invoked. Alternatively, use the 'Agg' backend (instead of 'QtAgg') above.
from networkx.algorithms.shortest_paths.weighted import dijkstra_path
import numpy

import pyjamas.rimage.rimutils as rimutils


# SHOULD THIS BE CALLED RImageOps and all methods be static (similar to PIL.ImageOps)
class rimage:
    """
    A class to do image operations.

    The image is stored as a numpy array (image_array).

    Because numpy arrays are accessed by row and col, all methods implemented here will use the row, col convention even to
    represent points.
    """

    image_extensions: Tuple[str] = ('.tif', '.tiff', '.jpg', '.jpeg', '.gif', '.png', '.bmp')
    livewire_shortest_path_fns: dict = {'A*': rimutils.rimutils.path_astar, 'Dijkstra': rimutils.rimutils.path_dijkstra}

    def __init__(self, image_in):
        self.image_array: numpy.array = image_in

    # Return a tuple with all the points between thesrc and thedst. thesrc and thedst are in (r, c) format
    def livewire(self, thesrc, thedst, margin, xy=False, smooth_factor=0.0, shortest_path_fn: Optional=rimutils.rimutils.path_dijkstra):
        # Copy parameters. This is critical. I had a headache, because the function was modifying the parameter values.
        src = thesrc.copy()
        dst = thedst.copy()

        # Crop image. Note that image_array.shape returns size in (rows, columns) format. But src and dst are in (col=x, row=y) format.
        minr = max(0, min(src[0], dst[0]) - margin)
        maxr = min(self.image_array.shape[0] - 1, max(src[0], dst[0]) + margin) + 1
        minc = max(0, min(src[1], dst[1]) - margin)
        maxc = min(self.image_array.shape[1] - 1, max(src[1], dst[1]) + margin) + 1

        # Crop image.
        if self.image_array.ndim == 3:
            im = self.image_array[0, minr:maxr, minc:maxc].copy()
        else:
            im = self.image_array[minr:maxr, minc:maxc].copy()

        # Adjust source and destination coordinates.
        src[0] -= minr
        src[1] -= minc
        dst[0] -= minr
        dst[1] -= minc

        # Make sure the cropped image is a matrix and invert it so that high intensities become low ones, and we can
        # search for the minimal cost path.
        im[:] = numpy.max(im) - im[:]  # This modifies the array in place, which avoids having to reallocate the memory
        # (which is slow).
        # Doing im = numpy.max(im) - im produces the same result, but reallocates the memory.

        # Create graph.
        G = rimutils.rimutils.makeGraphX(rimutils.rimutils.gaussian_smoothing(im, smooth_factor))

        # Find shortest path.
        path = shortest_path_fn(G, tuple(src), tuple(dst))

        # Generate coordinates.
        if xy:
            return [[c + minc, r + minr] for (r, c) in path]
        else:
            return [[r + minr, c + minc] for (r, c) in path]
